/* //////////////////////////////////////////////////////////
 * Music Player using the ESP8266-12
 * and the DFPlayer Mini MP3 Player
 * 
 * An access point (AP) and station (STA) webserver
 * to control the music player
 *
 * Connections:  
 * - DFPlayer RX (pin 2) to ESP8266 TX pin 5 (pin 14 also works)
 * - DFPlayer TX (pin 3) to ESP8266 RX pin 4 (pin 16 also works. pin 13 doesn't work)
 * - DFPlayer Busy (pin 16) to ESP8266 pin 12
 * - DFPlayer speaker output (pins 6,8) to a speaker  
 * - DFPlayer VCC (pin 1) to 5V
 * - DFPlayer GND (pin 7 or 10) to GND of the power supply and to Arduino GND
 * 
 * 
 * If the DFPlayer produces noise try one of the following (usually needed for the 5V Arduino):
 * - Lower the volume. Volume too high will cause noise and may crash the player
 * - Connect the RX pin in series with 1K resistors to the Arduino TX pin 
 * - Connect VCC (pin 1) in series with a 1N4001 diode to 5V (positive end of diode to 5V) to reduce the input voltage
 *   
 *   
 * SD card content:
 * - three folders: 01, mp3 and advert
 * - three files:
 *   - 01/001.mp3        for the playTrack(1, 1) call
 *   - mp3/0001.mp3      for the mp3.playMP3Folder(1) call
 *   - advert/0001.mp3   for the mp3.playAdvertFolder(1) call
 * - can have more folders from 01 to 99
 * - each folder can have more tracks from 001 to 255
 *   track names can be long filenames after the three digits
 *
 *
 * After uploading this program to the ESP8266-12 module,
 * open the serial monitor and restart the ESP8266 module to 
 * see what IP address is assigned to this webserver,
 * then browse to this IP address. For example
 *     192.168.1.99            for the main webpage
 *   or
 *     192.168.1.99/setup      for the WiFi setup page
 * 
 * or
 * 
 * Method 2 for after the program is up and running.
 * Connect your WiFi to this access point with the default ssid 
 *     "WiFi setup ESP8266-xxxx"
 * or the one you provided, then browse to the IP address
 *     1.2.3.4                  for the main webpage
 *   or
 *     1.2.3.4/setup            for the WiFi setup page
 * 
 * If the WiFi is not connected then you will have to
 * use method 2 above to connect your WiFi to the access point
 * and then browse to 1.2.3.4/setup to setup the ssid and password.
 * Alternatively, you can specify the ssid and password to connect to
 * in the setupWiFi(ssid, password) function call.
 * 
 * If using the ESP8266-01 module, nothing will be printed
 * on the serial monitor because it conflicts with 
 * the built-in led on pin 1
 * 
 * Copyright (c) 2016, 2020 Enoch Hwang
 */
                      
//////////////////////////////////////////////////////////////
//// VCC voltage stuff
ADC_MODE(ADC_VCC);

//////////////////////////////////////////////////////////////
//// status LED stuff
#define statusLed 2     // for ESP8266-12

//////////////////////////////////////////////////////////////
//// NTP clock stuff
#include "DateTime_RobotsForFun.h"

//////////////////////////////////////////////////////////////
//// WiFi stuff
#include "WiFi_RobotsForFun.h"

// Do delay without using the delay() function
void nonblockingDelay(int wait) {
  unsigned long timesUp;
  timesUp = millis() + wait;
  while (millis() < timesUp) {
    flashActivityLed();
    server.handleClient();
    yield();  // must have this to give time to handle wifi background stuff
  }
}


//////////////////////////////////////////////////////////////
// DFPlayer Mini MP3/WAV Player Stuff
#define DFPserial_RX 5  // DFP RX pin 2 connect to ESP8266 TX pin 5
#define DFPserial_TX 4  // DFP TX pin 3 connect to ESP8266 RX pin 4
//#define DFPserial_RX 14   // DFP RX pin 2 connect to ESP8266 TX pin 14
//#define DFPserial_TX 16   // DFP TX pin 3 connect to ESP8266 RX pin 16
#define BUSYPIN 12
#include "DFPlayer.h"
DFPlayer mp3;
int volume;

void setupDFPlayer() {
  mp3.begin();
  volume = 16;  // 30=loudest (too loud will make it not work)
  mp3.setVolume(volume);
  mp3.playTrack(1);
}

// If currently playing then need to play from the Advert folder
// otherwise need to play from the MP3 folder
// It is because it will not play from the Advert folder if no song is currently playing
void playMP3AdvertFolder(int track) {
  Serial.print("Playing track ");
  Serial.print(track);
  if (mp3.playing(BUSYPIN)) {
    Serial.println(" in Advert folder");
    mp3.playAdvertFolder(track);         // play given song in MP3 folder
  } else {
    Serial.println(" in MP3 folder");
    mp3.playMP3Folder(track);         // play given song in MP3 folder
  }
}

//////////////////////////////////////////////////////////////
// Alarm Clock Stuff
int alarmFolder, alarmTrack, alarmHour, alarmMinute;
void checkAlarmClock() {
  if (hour() == alarmHour && minute() == alarmMinute && second() == 0) {
    Serial.println("Alarm ringing");
    if (alarmFolder > 0 && alarmTrack > 0) {
      mp3.playTrack(alarmFolder, alarmTrack);
    } else if (alarmFolder > 0 && alarmTrack == 0) {
      mp3.playFolder(alarmFolder);
    } else if (alarmFolder == 0 && alarmTrack > 0) {
      mp3.playTrack(alarmTrack);
    }
  }
}


//////////////////////////////////////////////////////////////
//// HTML webpage
void indexHTML() {
  static String alarmSetString = "";
  static String m = "";
  // to get fast responsive webpage the code needs to execute the
  //   server.send(200, "text/html", msg);
  // command in this function as soon as possible

  // get url arguments and process the command here
  // do it here to see the changes reflect in the webpage being sent back to the client
  // the code here should not have long delay times otherwise the webpage will be very slow to load
//  if (server.hasArg("LED")) {
//    if (String(server.arg("LED")) == "OFF") {
//      digitalWrite(statusLed, LEDOFF);  // turn off LED
//      ledStatus = LEDOFF;
//      Serial.println("turn off LED");
//    } else if (String(server.arg("LED")) == "ON") {
//      digitalWrite(statusLed, LEDON);   // turn on LED
//      ledStatus = LEDON;
//      Serial.println("turn on LED");
//    } else if (String(server.arg("LED")) == "BLINK") {
//      ledStatus = LEDBLINK;                // blink LED
//      Serial.println("blink LED");
//    }
//  }

  if (server.argName(2).equals("PLAY")) {
    int folder = server.arg(0).toInt();
    int track = server.arg(1).toInt();
    if (folder > 0 && track > 0) {
      Serial.print("Playing track ");
      Serial.print(track);
      Serial.print(" in folder ");
      Serial.println(folder);
      m = "Playing track ";
      m += track;
      m += " in folder ";
      m += folder;
      mp3.playTrack(folder, track); // play given song in given folder
    } else if (folder > 0 && track == 0) {
      Serial.print("Playing all tracks in folder ");
      Serial.println(folder);
      m = "Playing all tracks in folder ";
      m += folder;
      mp3.playFolder(folder);       // play all songs in given folder
    }
  } else if (server.argName(1).equals("PLAY")) {
    int track = server.arg(0).toInt();
    if (track == 0) track = 1;
    playMP3AdvertFolder(track);
  } else if (server.arg(1).equals("UP")) {
    mp3.increaseVolume();
    volume = mp3.getVolume();
    Serial.print("Volume Up ");
    Serial.println(volume);
  } else if (server.arg(1).equals("DOWN")) {
    mp3.decreaseVolume();
    volume = mp3.getVolume();
    Serial.print("Volume Down ");
    Serial.println(volume);
//  } else if (server.argName(1).equals("EQ")) {
//    static int eq = 0;
//    eq++;
//    if (eq == 6) eq = 0;
//    mp3.setEQ(eq);
//    Serial.println("Eq");
  } else if (server.argName(1).equals("PREVIOUS") || server.argName(2).equals("PREVIOUS")) {
    mp3.previousTrack();
    Serial.println("Previous");
  } else if (server.argName(1).equals("STOP") || server.argName(2).equals("STOP")) {
    mp3.stopPlay();
    m = "";
    Serial.println("Stop");
  } else if (server.argName(1).equals("NEXT") || server.argName(2).equals("NEXT")) {
    mp3.nextTrack();
    Serial.println("Next");
  } else if (server.arg(3).equals("Set")) {
    alarmFolder = 0;
    alarmTrack = server.arg(0).toInt();
    if (alarmTrack == 0) alarmTrack = 1;
    alarmHour = server.arg(1).toInt();
    alarmMinute = server.arg(2).toInt();
    String aHString, aMString;
    if (alarmHour < 10) aHString = "0"+String(alarmHour);
    else aHString = String(alarmHour);
    if (alarmMinute < 10) aMString = "0"+String(alarmMinute);
    else aMString = String(alarmMinute);
    alarmSetString = "Alarm set: " + aHString + ":" + aMString + " global track " + String(alarmTrack) + "<br>";
    Serial.print("Alarm set: ");
    Serial.print(aHString);
    Serial.print(":");
    Serial.print(aMString);
    Serial.print(" track=");
    Serial.println(alarmTrack);
  } else if (server.arg(4).equals("Set")) {
    alarmFolder = server.arg(0).toInt();
    alarmTrack = server.arg(1).toInt();
    alarmHour = server.arg(2).toInt();
    alarmMinute = server.arg(3).toInt();
    String aHString, aMString;
    if (alarmHour < 10) aHString = "0"+String(alarmHour);
    else aHString = String(alarmHour);
    if (alarmMinute < 10) aMString = "0"+String(alarmMinute);
    else aMString = String(alarmMinute);
    alarmSetString = "Alarm set: " + aHString + ":" + aMString + " folder " + String(alarmFolder) + " track " + String(alarmTrack) + "<br>";
    Serial.print("Alarm set: ");
    Serial.print(aHString);
    Serial.print(":");
    Serial.print(aMString);
    Serial.print(" folder=");
    Serial.print(alarmFolder);
    Serial.print(" track=");
    Serial.println(alarmTrack);
  } else if (server.arg(3).equals("Clear") || server.arg(4).equals("Clear")) {
    alarmHour = -1;
    Serial.println("Alarm clear");
    alarmSetString = "";
  }

  // construct the webpage file using HTML coding
  String msg = "";
  msg += "<!DOCTYPE html>";
  msg += "<head>";
  msg += "<meta name=\"viewport\" content=\"width=device-width, initial-scale=1\">"; // keep fonts consistent between devices
  msg += "<title>Music Player</title>";
  msg += "  <style>";
  msg += "    body { background-color: #ffffff; font-family: Arial, Helvetica, Sans-Serif; Color: #000088; }";
  msg += "  </style>";
  msg += "</head>";
  
  msg += "<body>";
  msg += "<center><font size = \"3\"><b>Music Player</b></font><br>";
  msg += formattedTime();
  msg += "</center>";
  msg += "<hr><br>";

  // create user interface
  msg += m;
  msg += "  <form style=\"font-size:12pt\">";
  msg += "  <select name=\"folder\" style=\"width:345px; font-size:12pt\">";
  msg += "    <option selected disabled hidden>Select CD</option>";
  msg += "    <option value=\"01\">Folder 01</option>";
  msg += "    <option value=\"02\">Folder 02</option>";
  msg += "    <option value=\"03\">Folder 03</option>";
  msg += "    <option value=\"04\">Folder 04</option>";
  msg += "  </select>";
  
  msg += "<br><br>";
  msg += "Track number:&nbsp; <input type=\"number\" name=\"track\" min=\"1\" max=\"255\" style=\"font-size:14pt\" >";
  msg += "&nbsp;&nbsp;&nbsp;&nbsp;<input type=\"submit\" name=\"PLAY\" value=\"Play\" style=\"font-size:14pt\" >";

  msg += "<br><br>";
  msg += "Volume (";
  msg += volume;
  msg += ")&nbsp;:&nbsp; ";
//  msg += "<input type=\"submit\" name=\"VOLUME\" value=\"Loud\" style=\"font-size:14pt\" >";
//  msg += "&nbsp;&nbsp;&nbsp;<input type=\"submit\" name=\"VOLUME\" value=\"Soft\" style=\"font-size:14pt\" >";
  msg += "  <input type=\"submit\" name=\"VOLUME\" value=\"UP\" style=\"font-size:14pt\" >";
  msg += "&nbsp;&nbsp;<input type=\"submit\" name=\"VOLUME\" value=\"DOWN\" style=\"font-size:14pt\" >";

  msg += "<br><br>&nbsp;&nbsp;&nbsp;";
  msg += "<input type=\"submit\" name=\"PREVIOUS\" value=\"Previous\" style=\"font-size:14pt\" >";
  msg += "&nbsp;&nbsp;&nbsp;";
  msg += "<input type=\"submit\" name=\"STOP\" value=\"Stop\" style=\"font-size:14pt\" >";
  msg += "&nbsp;&nbsp;&nbsp;";
  msg += "<input type=\"submit\" name=\"NEXT\" value=\"Next\" style=\"font-size:14pt\" >";

  msg += "<br><hr><br>Alarm";
  msg += alarmSetString;
  msg += "Hour:&nbsp;<input type=\"number\" name=\"hour\" min=\"0\" max=\"23\" style=\"font-size:14pt\" >";
  msg += "&nbsp;&nbsp;&nbsp;Minute:&nbsp;<input type=\"number\" name=\"minute\" min=\"0\" max=\"59\" style=\"font-size:14pt\" >";
  msg += "<br><br>&nbsp;&nbsp;&nbsp;";
  msg += "<input type=\"submit\" name=\"ALARM\" value=\"Set\" style=\"font-size:14pt\" >";
  msg += "&nbsp;&nbsp;&nbsp;<input type=\"submit\" name=\"ALARM\" value=\"Clear\" style=\"font-size:14pt\" >";
  msg += "</form>";
  
  // wifi setup and info links
  msg += "<br><hr>IP address: ";
  if (wifiIsConnected) {
    msg += "<a href=\"/\">" + WiFi.localIP().toString() + "</a>";
    msg += "<br>Wifi setup and info: ";
    msg += "<a href=\"/setup\">" + WiFi.localIP().toString() + "/setup</a>";
  } else {
    msg += "<a href=\"http://1.2.3.4\">1.2.3.4</a>";
    msg += "<br>Wifi setup and info: ";
    msg += "<a href=\"http://1.2.3.4/setup\">1.2.3.4/setup</a>";
  }
  msg += "</body>";
  msg += "</html>";
  // end of webpage file

  // send webpage to client browser
  server.send(200, "text/html", msg);

  // get url arguments and process the command here
  // changes here will not be reflected in the webpage being sent (because it has already been sent)
  // therefore the code here can have long delay times since the webpage has already been sent back to the client

}

void setup() {
  Serial.begin(115200);
  pinMode(statusLed, OUTPUT);
  pinMode(BUSYPIN, INPUT);

  char devicename[] = "Music Player Webserver"; // optional. change to whatever you want. Use "ping ESP8266" to test
  // setup an access point (AP) for devices to connect to this ESP8266 directly without needing a wifi router
  setupAP(devicename);  // setup access point with the devicename as the ssid and with no password

  // use DHCP, default device name, ssid and password obtained from EEPROM, and default blocking if cannot connect to wifi
  setupWiFi(devicename);
  
  // set clock to Pacific Standard Time
  // it will also automatically adjust for Daylight Savings Time
  setClock(PST);
                  
  // update systemStartTime and systemStartCount. This is for information purposes only
  systemStartTime = now();
  EEPROM.get(EEPROM_ADDRESS_systemStartCount, systemStartCount);  // get systemStartCount from EEPROM
  systemStartCount++;
  EEPROM.put(EEPROM_ADDRESS_systemStartCount, systemStartCount);
  EEPROM.commit();

  // MP3 Player stuff
  setupDFPlayer();    
}


void loop() {

  if (APisOn && (millis() > 600000)) turnOffAP(); // turn off AP after 10 minutes
  flashActivityLed();     // show system is alive
  server.handleClient();  // handle web client request
}
